#include <AppKit.h>
#include <InterfaceKit.h>
#include <KernelKit.h>
#include <StorageKit.h>
#include "ikbd.h"
#include <stdio.h>
#include "config.h"
#include "defs.h"

#define MSG_REDRAW 1

extern "C"
{
	extern int scr_width;
	extern int scr_height;

	void OpenWindow();
	void CloseWindow();
	void ResizeWindow();
	void Update();

	void InitApp(int argc, char *argv[]);
	void DeleteApp(void);

	extern int main_beos(int argc, char *argv[]);
	
	extern uchar *bits;
	extern ulong xmod;
	extern int shiftmod;
	extern void ikbd_button (int button, int pressed);
	extern void ikbd_pointer (int x, int y, int max_x, int max_y);
	extern void Lock();
	extern void Unlock();
	extern void DirectDraw(int x1, int y1, int x2, int y2);
	extern void mouse_update();
	extern void keyboard_update();
	extern void ikbd_key (int keycode, int pressed);
	extern int mb,mx,my;
	extern uchar *kstates;
	extern char mapcol[];
	extern char glbquit;
	extern unsigned short old_color[16];
	extern void change_colors (void);
	extern void kill_beos();
	extern void execute_start (unsigned long new_pc);
	extern unsigned long ExpTabHi[];
	extern char *AllocMem(ulong size);
	extern void FreeMem(char *mem);
};

class		HPView;
class		HPWindow;

HPView		*the_view;
HPWindow 	*the_window;
BBitmap		*the_bitmap;
uchar		*bits;
ulong		xmod;
uchar 		*kstates;
thread_id	tg;

static bool udp=TRUE;

void change_colors (void)
{	int i;
	unsigned long pixels[16];

	if (shiftmod == 2)
	{
	}
	else for (i=0; i<16; i++)
	{	int r,g,b,c;
		c = (i < (shiftmod == 0 ? 16 : 4) ? old_color[i] : 0);
#if !STE
		c &= 0x777;
#endif

		r	= (((c>>8)&7)<<5)	| ((c>> 7)&16) | 0x0F;
		g	= (((c>>4)&7)<<5)	| ((c>> 3)&16) | 0x0F;
		b	= (((c   )&7)<<5)	| ((c<< 1)&16) | 0x0F;

		mapcol[i]=index_for_color(r,g,b);
	}
}

void AllocBitmap()
{
	switch(shiftmod)
	{
		case	0:
				the_bitmap = new BBitmap(BRect(0, 0, scr_width-1, scr_height-1),B_COLOR_8_BIT);
				break;
		case	1:
				the_bitmap = new BBitmap(BRect(0, 0, scr_width-1, (scr_height*2)-1),B_COLOR_8_BIT);
				break;
		case	2:
				the_bitmap = new BBitmap(BRect(0, 0, scr_width-1, scr_height-1),B_COLOR_8_BIT);
				break;
	}
	bits=(uchar *)the_bitmap->Bits();
	xmod=the_bitmap->BytesPerRow();
}

class HPView : public BView
{
	char	cursor[68];
	
	public:
			HPView(void):BView(BRect(0, 0, scr_width-1, scr_height-1),"", B_FOLLOW_ALL,B_WILL_DRAW)
			{
				SetDrawingMode(B_OP_COPY);
				SetHighColor(0,0,0);
				SetLowColor(255,255,255);
				cursor[0]=16;
				cursor[1]=01;
				cursor[2]=00;
				cursor[3]=00;
				for(int i=4;i<68;i++)cursor[i]=0;
			};
			
	void 	MouseMoved(BPoint pos, ulong transit, BMessage *msg)
			{
				if(transit==B_ENTERED_VIEW)
				{
					be_app->SetCursor(cursor);
				}
				if(transit==B_EXITED_VIEW)
				{
					be_app->SetCursor(B_HAND_CURSOR);
				}
			};

	void	Draw(BRect rect)
			{
				DrawBitmap(the_bitmap,BPoint(0,0));
				Sync();
			};
};

class HPWindow : public BWindow 
{
	public:
	
			HPWindow() : BWindow(BRect(0, 0, scr_width-1, scr_height-1), "STonBeOS by Tinic", B_TITLED_WINDOW, B_NOT_ZOOMABLE | B_NOT_RESIZABLE)
			{
				SetDiscipline(FALSE);
				MoveTo(80, 60);
				Lock();
				the_view = new HPView();
				AddChild(the_view);
				Unlock();

				for (int i=0; i<16; i++)
				{
					mapcol[i]=i<<1;
				}
				
				AllocBitmap();
				
				the_view->MakeFocus();
				
				Show();
			};
			
			bool QuitRequested()
			{
				Lock();
				udp=FALSE;
				kill_thread(tg);
				kill_beos();
				be_app->PostMessage(B_QUIT_REQUESTED);
				return TRUE;
			};
			
			void WindowActivated(bool state)
			{
				udp=state;
			};			
};

static int argc_;
static char **argv_;

class HPApp : public BApplication
{
	public:
	
			HPApp():BApplication('ATRI')
			{
			};
			
	void	ReadyToRun()
			{
				BDirectory *app_dir;
				app_info the_info;
				app_dir = new BDirectory;
				char app_path[1024];

				if (be_app->GetAppInfo(&the_info) == B_NO_ERROR) 
				{
					BFile *file = new BFile;
	
					file->SetRef(the_info.ref);
					file->GetParent(app_dir);
					app_dir->GetPath(app_path,1024);
					chdir(app_path);

					delete file;
				}
				delete app_dir;
				main_beos(argc_,argv_);
				resume_thread(tg=spawn_thread((thread_entry)execute_start,"MC68000",B_NORMAL_PRIORITY,(void *)TOSSTART));
			};
};

void InitApp(int argc, char **argv)
{
	argc_=argc;
	argv_=argv;
	
	be_app=(BApplication *)new HPApp();
	be_app->Run();
}


void Lock()
{
	the_window->Lock();
}

void Unlock()
{
	the_view->Sync();
	the_window->Unlock();
}

void DirectDraw(int x1, int y1, int x2, int y2)
{
	BRect r1=BRect(x1,y1,x2-1,y2-1);
	the_view->DrawBitmapAsync
	(
		the_bitmap,
		r1,
		r1
	);
}

int mb=0,mx=0,my=0;

void mouse_update()
{
	if(!udp)return;
	ulong b;
	BPoint p;
	the_window->Lock();
	the_view->GetMouse(&p,&b);
	the_window->Unlock();
	if(shiftmod==1)
	{
		p.y/=2;
	}
	if(p.x<0)p.x=0;
	if(p.x>scr_width)p.x=scr_width;
	if(p.y<0)p.y=0;
	if(p.y>scr_height)p.y=scr_height;
	mx=p.x;
	my=p.y;
	mb=b&3;
}

void keyboard_update()
{
	static key_info ki;
	kstates=ki.key_states;	
	if(!udp)
	{
		for(int i=0;i<16;i++)
		{
			kstates[i]=0;
		}
		return;
	}
	the_window->Lock();
	the_view->GetKeys(&ki,FALSE);
	the_window->Unlock();
}

#define CHUNK_LINES 16

void Update()
{	
	the_window->Lock();
	for (int i=0; i<(scr_height/CHUNK_LINES); i++)
		DirectDraw
		(
			0, 		 	(i+0)*CHUNK_LINES,
			scr_width,	(i+1)*CHUNK_LINES
		);
	the_view->Sync();
	the_window->Unlock();
}

void ResizeWindow()
{
	delete the_bitmap;
	AllocBitmap();
	switch(shiftmod)
	{
		case	0:
		case	2:
				the_window->ResizeTo(scr_width, scr_height);
				break;
		case	1:
				the_window->ResizeTo(scr_width, scr_height*2);
				break;
	}
}

void DeleteApp()
{
	delete be_app;
}

void OpenWindow()
{
	for (int i=0; i<0x100; i++) 
	{
		ExpTabHi[(i*2)] = ((i & 0x80) ? 0 : 0xff000000)
					| ((i & 0x40) ? 0 : 0x00ff0000)
					| ((i & 0x20) ? 0 : 0x0000ff00)
					| ((i & 0x10) ? 0 : 0x000000ff);
		ExpTabHi[(i*2)+1] = ((i & 0x08) ? 0 : 0xff000000)
					| ((i & 0x04) ? 0 : 0x00ff0000)
					| ((i & 0x02) ? 0 : 0x0000ff00)
					| ((i & 0x01) ? 0 : 0x000000ff);
	}
	the_window = new HPWindow();
}

void CloseWindow()
{
}

char *AllocMem(ulong size)
{
	return new char[size];
}

void FreeMem(char *mem)
{
	delete [] mem;
}
