/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 
/* When the bug in gcc 2.6.0-2.6.3 is fixed, define as 'static' ... */
#define STATIC 

#define DEBUG_EXCEPT 0
#define DEBUG_TRAPS 0
#define DEBUG_MEMACC 0

#include <stdio.h>
#include <string.h>
#include "missing.h"
#if 0
#include <setjmp.h>
#endif
#include "defs.h"
#include "cpu.h"
#include "debug.h"
#include "am.h"
#include "main.h"
#include "mem.h"
#include "onexit.h"
#include "prof.h"
#include "screen.h"
#include "xlib_vdi.h"
#include "gemdos.h"

#if MONITOR
extern int in_monitor;
#endif
#define CHECK_TRACE() do{if (sr&SR_T) {flags &= ~F_TRACE1; flags |= F_TRACE0;}}while(0)
#define CHECK_NOTRACE() do{if ((sr&SR_T)==0) flags &= ~(F_TRACE0|F_TRACE1);}while (0)
#define SBITS(_l,_h,_x)	((int) ((((_x)>>(_l))^(1<<((_h)-(_l))))&\
						((1<<((_h)-(_l)+1))-1)) - (1<<((_h)-(_l))))
#define UBITS(_l,_h,_x) ((((unsigned int)(_x))>>(_l))&((1<<((_h)-(_l)+1))-1))
#define BIT(_n,_x)	(((_x)>>(_n))&1)

UL dummy0[100];
L dreg[17];

#if !defined(PCREG) || !defined(__GNUC__)
L pc;
#endif

#if !defined(SRREG) || !defined(__GNUC__)
UW sr;	/* includes CCR */
#endif

#if !defined(NZREG) || !defined(__GNUC__)
L nz_save;
#endif

UL flags=0;
UW cw;
UL shadow_pc;

/* these should be global register variables... */
UL vaddr;
void *addr;
extern B *mem;
/* tables for handling special addresses */
#include "iotab1.c"
#include "iotab2.c"

static INLINE B LOAD_TABLE1(UL s)
{
	UB x = IOTAB1_flags_LOAD[s-TRIM(IOTAB1_LO)];
	if (x == 0) BUS_ERROR(s, 1);
	else if (x == 1) return LM_B(ADDR(s));
	else return IOTAB1_funcs_LOAD[x-2]();
}


static INLINE B LOAD_TABLE2(UL s)
{
	UB x = IOTAB2_flags_LOAD[s-TRIM(IOTAB2_LO)];
	if (x == 0) BUS_ERROR(s, 1);
	else if (x == 1) return LM_B(ADDR(s));
	else return IOTAB2_funcs_LOAD[x-2]();
}




static INLINE void STORE_TABLE1(UL d, B v)
{
	UB x = IOTAB1_flags_STORE[d-TRIM(IOTAB1_LO)];
	if (x == 0) BUS_ERROR(d, 0);
	else if (x == 1) SM_B(ADDR(d), v);
	else IOTAB1_funcs_STORE[x-2](v);
}


static INLINE void STORE_TABLE2(UL d, B v)
{
	UB x = IOTAB2_flags_STORE[d-TRIM(IOTAB2_LO)];
	if (x == 0) BUS_ERROR(d, 0);
	else if (x == 1) SM_B(ADDR(d), v);
	else IOTAB2_funcs_STORE[x-2](v);
}


/* The following few functions handle I/O and other 'special' memory accesses
 */
B LS_B (UL s)
{
	if (!SUPERVISOR_MODE) BUS_ERROR (s, 1);
#if PROTECT_ROM
#if SMALL
#if TOS_1
	else if (s > 0x800 && s < TRIM(CARTSTART))
#else
	else if (s > 0x800 && s < TRIM(TOSSTART))
#endif
	{
		BUS_ERROR(s, 1);
	}
#endif 
#if TOS_1
	else if (s < TRIM(CARTEND))
#else
	else if (s < TRIM(TOSEND) || (s >= TRIM(CARTSTART) && s < TRIM(CARTEND)))
#endif
		return LM_B(ADDR(s));
#endif
	else if (s < TRIM(IOTAB1_LO)) BUS_ERROR (s, 1);
#if DEBUG_MEMACC
	{
#endif
	if (s <= TRIM(IOTAB1_HI))
	{
		return LOAD_TABLE1(s);
	}
	else if (s >= TRIM(IOTAB2_LO) && s <= TRIM(IOTAB2_HI))
	{
		return LOAD_TABLE2(s);
	}
	else BUS_ERROR (s, 1);
#if DEBUG_MEMACC
	}
#endif
}

W LS_W (UL s)
{
	if (s & 1) ADDRESS_ERROR (s, 1);
	else
	{
#if PROTECT_ROM
#if SMALL
#if TOS_1
		if (s > 0x800 && s < TRIM(CARTSTART-1))
#else
		if (s > 0x800 && s < TRIM(TOSSTART-1))
#endif
		{
			BUS_ERROR(s, 1);	/* bus err or addr. err 1st?*/
		}
		else
#endif
#if TOS_1
		if (s < TRIM(CARTEND-1))
#else
		if (s < TRIM(TOSEND-1) || (s >= TRIM(CARTSTART) && s < TRIM(CARTEND-1)))
#endif
			return LM_W(ADDR(s));
		else
#endif
		if (s < TRIM(IOTAB1_LO)) BUS_ERROR (s, 1);
		else
		{
		return (LS_B(s)<<8)|(LS_B(s+1) & 0xff);
		}
	}
}

L LS_L (UL s)
{
	if (s & 1) ADDRESS_ERROR (s, 1);
	else
	{
#if PROTECT_ROM
#if SMALL
#if TOS_1
		if (s > 0x800 && s < TRIM(CARTSTART-3))
#else
		if (s > 0x800 && s < TRIM(TOSSTART-3))
#endif
			BUS_ERROR(s, 1);	/* bus err or addr. err 1st?*/
		else
#endif
#if TOS_1
		if (s < TRIM(CARTEND-3))
#else
		if (s < TRIM(TOSEND-3) || (s >= TRIM(CARTSTART) && s < TRIM(CARTEND-3)))
#endif
			return LM_L(ADDR(s));
		else
#endif
		if (s < TRIM(IOTAB1_LO)) BUS_ERROR (s, 1);
		else
		{
		return (LS_W(s)<<16)|(LS_W(s+2)&0xffff);
		}
	}
}

void SS_B (UL d, B v)
{
	if (!SUPERVISOR_MODE) BUS_ERROR (d, 0);
	if (d < 0x800) SM_B(ADDR(d), v);
	else if (d < TRIM(IOTAB1_LO)) /* ROM */
	{
		BUS_ERROR (d, 0);
	}
	else if (d <= TRIM(IOTAB1_HI)) STORE_TABLE1(d, v);
	else if (d >= TRIM(IOTAB2_LO) && d <= TRIM(IOTAB2_HI)) STORE_TABLE2(d, v);
	else BUS_ERROR (d, 0);
}

void SS_W (UL d, W v)
{
	if (d & 1) ADDRESS_ERROR (d, 0);
	else
	{
		if (d < 0x800 && SUPERVISOR_MODE) SM_W(ADDR(d), v);
		else
		{
			SS_B(d, v>>8);
			SS_B(d+1,v&0xff);
		}
	}
}

void SS_L (UL d, L v)
{
	SS_W(d, v>>16);
	SS_W(d+2, v&0xffff);
}

static INLINE void SAVE_ADDRESS (void *d)
{
	addr = d;
}
static INLINE B CLS_B (UL s)
{
	addr = 0;
	vaddr = s;
	return CL_B(s);
}
static INLINE UB CLS_UB (UL s)
{
	addr = 0;
	vaddr = s;
	return CL_UB(s);
}
static INLINE W CLS_W (UL s)
{
	addr = 0;
	vaddr = s;
	return CL_W(s);
}
static INLINE UW CLS_UW (UL s)
{
	addr = 0;
	vaddr = s;
	return CL_UW(s);
}
static INLINE L CLS_L (UL s)
{
	addr = 0;
	vaddr = s;
	return CL_L(s);
}
static INLINE UL CLS_UL (UL s)
{
	addr = 0;
	vaddr = s;
	return CL_UL(s);
}

#if 1
#define US_B(x) \
	if (!addr) CS_B (vaddr, x); \
	else SR_B(addr, x);
	
#define US_UB(x) \
	if (!addr) CS_UB (vaddr, x); \
	else SR_UB(addr, x);

#define US_W(x) \
	if (!addr) CS_W (vaddr, x); \
	else SR_W(addr, x);
	
#define US_UW(x) \
	if (!addr) CS_UW (vaddr, x); \
	else SR_UW(addr, x);

#define US_L(x) \
	if (!addr) CS_L (vaddr, x); \
	else SR_L(addr, x);

#define US_UL(x) \
	if (!addr) CS_UL (vaddr, x); \
	else SR_UL(addr, x);

#else

static INLINE void US_B (B x)
{
	if (!addr) CS_B (vaddr, x);
	else SR_B(addr, x);
}
static INLINE void US_UB (UB x)
{
	if (!addr) CS_UB (vaddr, x);
	else SR_UB(addr, x);
}
static INLINE void US_W (W x)
{
	if (!addr) CS_W (vaddr, x);
	else SR_W(addr, x);
}
static INLINE void US_UW (UW x)
{
	if (!addr) CS_UW (vaddr, x);
	else SR_UW(addr, x);
}
static INLINE void US_L (L x)
{
	if (!addr) CS_L (vaddr, x);
	else SR_L(addr, x);
}
static INLINE void US_UL (UL x)
{
	if (!addr) CS_UL (vaddr, x);
	else SR_UL(addr, x);
}
#endif

/* ---- */
void enter_user_mode (void)
{
	L tmp;
	sr &= ~SR_S;
	tmp = AREG(7);
	AREG(7) = AREG(8);
	AREG(8) = tmp;
}

void enter_supervisor_mode (void)
{
	L tmp;
	sr |= SR_S;
	tmp = AREG(7);
	AREG(7) = AREG(8);
	AREG(8) = tmp;
}

static INLINE void V_EX (UL addr, int rw)
{
	UW tsr = sr;
	FIX_CCR();
	if (!SUPERVISOR_MODE)
		enter_supervisor_mode();
	flags &= ~(F_TRACE0|F_TRACE1); /* no Tracing during exceptions */
	sr &= ~SR_T;
	PUSH_UL(pc);
	PUSH_UW(tsr);
#if 1
	PUSH_UW(0);	/* should be cw */
	PUSH_UL(addr);
	PUSH_UW(tsr & 0xff00 | (rw ? 0x10 : 0)); /* BUG (ssr) */
#endif
#if 0
	sr &= 0xf8ff;
	sr |= 0x0700;	/* BUG */
#endif
}

void EX_BUS_ERROR (UL addr, int rw)
{
	UL new = LM_UL(ADDR(8));
#if 0
	fprintf(stderr,"Bus error at %lx\n",pc);
#endif
	V_EX (addr, rw);
}

void EX_ADDRESS_ERROR (UL addr, int rw)
{
	UL new = LM_UL(ADDR(12));
	V_EX (addr, rw);
}

void EXCEPTION (int n)
{
	UW tsr;
	UL new;
	FIX_CCR();
	tsr = sr;
	new = LM_UL(ADDR(n*4));
#if 0
	if (n!=28 && n!=69)
	fprintf(stderr,"Exception %d at pc=%08lx\n",n,pc);
#endif
	if (n < 32 || n >= 48)
	{
		flags &= ~(F_TRACE0|F_TRACE1); /* no Tracing during exceptions */
		sr &= ~SR_T;
	}
#if DEBUG_EXCEPT
#if DEBUG_TRAPS
	if (n >= 32 && n < 48) switch (n-32)
	{
		case 1: D("GEMDOS(%02x) -> %lx\n", LM_W(MEM(SP)), new); break;
		case 2: D("AES/VDI(%02x,%08x) -> %lx\n", DREG(0), DREG(1),new); break;
		case 13: D("BIOS(%02x) -> %lx\n", LM_W(MEM(SP)), new); break;
		case 14: D("XBIOS(%02x) -> %lx\n", LM_W(MEM(SP)) ,new); break;
		default: D("Trap #%d -> %lx\n", n-32, new); break;
	}
	else
#endif	
	{
	}
#endif
#if 1
	if (n == 34 && vdi_mode)
	{
#if REDRAW
		if (DREG(0) == 200 && redraw_flag)	/* AES */
		{
			/* window must be updated, pretend form_dial(FMD_FINISH,...) */
			new = 0xfa0100;
			redraw_flag = 0; /* important! */
		}
		else
#endif
		if (DREG(0) == 115)
		{
			Vdi(0);
			if (vdi_done) return;
		}
	}
#endif
	if (!SUPERVISOR_MODE)
		enter_supervisor_mode();
	PUSH_UL(pc);
	PUSH_UW(tsr);
	pc = new;	/* should CHECK_PC() here? */
}

void TRAP (int n)
{
	EXCEPTION(32+n);
}

void ILLEGAL (void)
{
	pc -= 2;
	EXCEPTION (T_ILLEGAL);
}

/* All BCD functions are UNTESTED! */
#define BCD2DEC(_x) ((((_x)&0xf0)>>4)*10+((_x)&0xf))
#define DEC2BCD(_x) ((((_x)/10)<<4)+((_x)%10))

B BCD_ADD(B a, B b)
{
	int q,w;
	sr &= ~MASK_CC|MASK_CC_X;
	q = BCD2DEC(a);
	w = BCD2DEC(b);
	w += q + GET_X_BIT();
	if (w >= 100)
	{
		w -= 100;
		sr |= MASK_CC_X|MASK_CC_C;
	}
	if (w != 0) UNSET_Z();	/* ???? */
	return DEC2BCD(w);
}

B BCD_SUB(B a, B b)
{
	int q,w;
	sr &= ~MASK_CC|MASK_CC_X;
	q = BCD2DEC(a);
	w = BCD2DEC(b);
	w -= q + GET_X_BIT();
	if (w < 0)
	{
		w += 100;
		sr |= MASK_CC_X|MASK_CC_C;
	}
	if (w != 0) UNSET_Z();	/* ???? */
	return DEC2BCD(w);
}

/* -------------------- SPECIAL EVENTS DISPATCHER --------------------------- */
#define MFP_ICOUNT_DEF 500
static int syst_count=0;
static int last_acia=0;
static int mfp_prio=0;
static int mfp_Timer_A_count=0;
static int mfp_icount=MFP_ICOUNT_DEF;
extern volatile UB mfp_TADR;
extern int mfp_TA_delay;
#if PROFILE
void process_flags (unsigned int iw)
{
	if (flags & F_PROFILE)
	{
		extern int prof_freq[];
		extern unsigned char prof_maptab[];
		prof_freq[prof_maptab[iw]]++;
	}
#else
void process_flags (void)
{
#endif
#if TIMER_A
	if ((flags & F_TIMERA_ON))
	{
		if (mfp_Timer_A_count)	/* need to generate Timer A interrupts */
		{
			mfp_Timer_A_count--;
#if 0
fprintf(stderr,"Timer A interrupt\n");
#endif
			if (IPL_OK(6))
			{
				EXCEPTION(T_TIMERA);
				SET_IPL(7);
				return;
			}
		}
		else if (--mfp_icount <= 0)
		{
			mfp_Timer_A_count = mfp_get_count();
			mfp_icount = MFP_ICOUNT_DEF;
		}
	}
#endif /* TIMER_A */
	if (flags & F_BLIT)
	{
		Do_Blit();
		flags &= ~F_BLIT;
	}
	else
	if ((flags & F_ACIA) && IPL_OK(6)
			/* && mfp_prio < 6 */ /* && syst_count != last_acia */)
	{
/*		mfp_prio = 6; */
		EXCEPTION(T_ACIA);
	/*	last_acia = syst_count; */
		SET_IPL(7);
		flags &= ~F_ACIA;
	}
	else
	if ((flags & F_RCV_FULL) && IPL_OK(6))
	{
		EXCEPTION(T_RCV_FULL);
		SET_IPL(7);
		flags &= ~F_RCV_FULL;
	}
	if (flags & F_TRACE0)
	{
		flags &= ~F_TRACE0;
		flags |= F_TRACE1;
	}
	else if (flags & F_TRACE1)
	{
		EXCEPTION(T_TRACE);
	}
#if MONITOR
	if (!in_monitor)
	{
#endif
#if DOUBLE_HZ200
	if ((flags & (F_200Hz|F_200HzB)) && (IPL_OK(6) /* && mfp_prio < 5 */
#else
	if ((flags & (F_200Hz)) && (IPL_OK(6) /* && mfp_prio < 5 */
#endif
			&& (LM_UB(MEM(0xfffa11)) & 0x20) == 0))
	{
		/* Set it -> Timer C interrupt is being processed */
		SM_UB(MEM(0xfffa11),LM_UB(MEM(0xfffa11)) | 0x20);
		EXCEPTION(T_200Hz);
		SET_IPL(7);
#if DOUBLE_HZ200
		if (flags & F_200Hz)
		{
			flags &= ~F_200Hz;
			flags |= F_200HzB;
		}
		else flags &= ~F_200HzB;
#else
		flags &= ~F_200Hz;
#endif
	}
	else if ((flags & F_VBL) && IPL_OK(4))
	{
		machine.screen_shifter();
		EXCEPTION(T_VBL);
		SET_IPL(5);
		flags &= ~F_VBL;
	}
#if MONITOR
	}
#endif
}

UB testcond[] = {
#include "testtab.c"
};

#if 0
#include "old.c"
#endif

#ifdef GEN_FUNCTAB
void Nil (unsigned int iw)
{
	ILLEGAL();
}
#ifdef DECOMPILER
#define CODE_FUNC static inline
#else
#define CODE_FUNC
#endif
#include "code.c"
#endif

int count=0;
#ifndef DECOMPILER
#if 1||MONITOR||!(__i486__ && linux)
void execute (UL new_pc)
{
#ifdef GEN_LABELTAB
#define CODE_LABEL_TABLE
#include "code.c"
#endif
	pc = new_pc;
	for (;;)
	{
		unsigned int iw;
#if MONITOR
		if (update_monitor(dreg, sr, pc) == 'q') return;
		count++;
#endif
#if PROFILE
		if (flags) process_flags(iw);
		iw=LM_UW(MEM(pc));
#else
		if (flags) process_flags();
		iw=LM_UW(MEM(pc));
#endif
		pc+=2;
#if defined(GEN_FUNCTAB)
		jumptab[iw](iw);
#elif defined(GEN_SWITCH)
#include "code.c"
#elif defined(GEN_LABELTAB)
#undef CODE_LABEL_TABLE
#include "code.c"
#endif
	}
}
#endif

void execute_start (UL new_pc)
{
	execute (new_pc);
}

#if MONITOR
void exfunc (int status)
{
	fprintf (stderr, "Instructions: %d\n", count);
}
#endif

void init_cpu (void)
{
	int i;
#if MONITOR
	OnExit (exfunc);
#endif
	bzero ((char *)dreg, 17*4);
	sr = 0x2700;
	SM_UL(ADDR(0),LM_UL(MEM(TOSSTART)));
	SM_UL(ADDR(4),LM_UL(MEM(TOSSTART+4)));
	SP = LM_UL(ADDR(0));
	pc = LM_UL(ADDR(4));
#if 0
	for (i=0xd000;i<0xe000;i++)
		jumptab[i]=L1101Code;
	for (i=0x0000;i<0x1000;i++)
		jumptab[i]=L0000Code;
	for (i=0x4000;i<0x5000;i++)
		jumptab[i]=L0100Code;
	for (i=0x5000;i<0x6000;i++)
		jumptab[i]=L0101Code;
	for (i=0xb000;i<0xc000;i++)
		jumptab[i]=L1011Code;
	for (i=0xc000;i<0xd000;i++)
		jumptab[i]=L1100Code;
#endif
#if 0
	for (i=0x8000;i<0x9000;i++)
		jumptab[i]=L1000Code;
	for (i=0x9000;i<0xa000;i++)
		jumptab[i]=L1001Code;
#endif
}
#endif
