/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 */
#include "defs.h"
#include "debug.h"
#include "prof.h"
#include "tosdefs.h"
#include <signal.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <KernelKit.h>

#define INTERRUPTS 1

/* main.c */
extern long tv_usec;
extern int hz200pt, vblpt, audio;
extern B *Image, *mem;
extern void show_screen (void);
extern char *serial_dev;

static int vblcount, systcount;
static thread_id ta;
static double usec;
sem_id a_sem;

/*--------- Information copied from hardware regs at regular intervals: ------*/
volatile UL vbase;
UL old_vbase=0;
/*------------ Information directly accessed though hw regs  -----------------*/
/*                   (watch out for unaligned accesses!!!)                    */
UW *color;
/*----------------------------------------------------------------------------*/

/* This is where all hardware emulation takes place
 */
void hardware_jobs (int dummy)
{
	while(acquire_sem(a_sem)==B_NO_ERROR)
	{
		if (audio) audio_update();
#if 1
		if (serial_dev != NULL && serial_stat())
		{
			SM_UB(MEM(0xfffa2b),LM_UB(MEM(0xfffa2b)) | 0x80);
			flags |= F_RCV_FULL;
		}
#endif
		if (--vblcount == 0)
		{
#if INTERRUPTS
			flags |= F_VBL;
#endif
			vblcount = vblpt;
			if (old_vbase != vbase)
			{
#if 0
				fprintf(stderr,"New screen address is %lx\n", vbase);
#endif
				old_vbase = vbase;
			}
		}
#if INTERRUPTS
		if (--systcount == 0)
		{
			systcount = hz200pt;
			if (LM_UB(MEM(0xfffa09)) & 0x20) flags |= F_200Hz;
		}
#endif
	}
}

void snoozer()
{
	while(1)
	{
		snooze((double)(usec));
		release_sem(a_sem);
	}
}

void crash (int sig)
{
//	fprintf (stderr, "Got a fatal signal %d [pc=%08lx]\n", sig, pc);
}

void init_hardware (void)
{
	color = (UW *)MEM(M_COLORS);
	
	usec = (double)tv_usec;
	vblcount = vblpt;
	systcount = hz200pt;

	a_sem=create_sem(0,"a_sem");
	resume_thread(spawn_thread((thread_entry)hardware_jobs,"hardware",B_NORMAL_PRIORITY,NULL));
	resume_thread(ta=spawn_thread((thread_entry)snoozer,"timer",B_NORMAL_PRIORITY,NULL));
	
	signal (SIGILL, crash);
}

void kill_hardware (void)
{
	kill_thread(ta);
	delete_sem(a_sem);
}

