/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 */
#include "defs.h"
#include "mem.h"
#include <stdio.h>

B *mem;

extern int warmboot,fix_screen,scr_height,scr_width,scr_planes;
extern volatile UL vbase;
extern char *AllocMem(unsigned long size);
extern void FreeMem(char *mem);

void init_mem (void)
{
	int mem_wanted;
	UL new_top;
	
	printf("Allocating %4.1f Mbytes of ST memory...\n",((float)MEMSIZE)/(1024.0*1024.0));
	
	if(!(mem = (B *)AllocMem(MEMSIZE)))
	{
		printf("Fatal: not enough memory!\n");
		exit(0);
	}
	else
	{
//		printf("Cleaning allocated memory...\n");
//		memset(mem,0,MEMSIZE);
	}
	
	if (fix_screen)
	{
#if SMALL
#define PHYSTOP 0x400000
#else
#define PHYSTOP 0xe00000
#endif
		mem_wanted = (scr_height*scr_width*scr_planes)/8;
		new_top = (PHYSTOP-mem_wanted)&-1024;
		SM_UL(MEM(0x420),0x752019f3);
		SM_UL(MEM(0x43a),0x237698aa);
		SM_UL(MEM(0x51a),0x5555aaaa);	/* might break older TOSes <1.02? */
		fprintf(stderr,"Initializing memory for custom screen size... (top=%lx)\n",new_top);
#if 0
		SM_UL(MEM(0x426),0x31415926);
		SM_UL(MEM(0x42a),0xfa1400);
#endif
		SM_UL(MEM(0x4ba),16000);
		SM_UL(MEM(0x436),new_top);
		SM_UL(MEM(0x42e),LM_UL(MEM(0x436)));
#if 0
		fprintf(stderr,"phystop is now %lx\n",LM_UL(MEM(0x42e)));
#endif
		SM_UL(MEM(0x44e),LM_UL(MEM(0x42e)));
#if SMALL
		SM_UB(MEM(0x424),10);
#else
		SM_UB(MEM(0x424),15);
#endif
		set_vbase(new_top);
	}
	else if (warmboot)
	{
		SM_UL(MEM(0x420),0x752019f3);
		SM_UL(MEM(0x43a),0x237698aa);
		SM_UL(MEM(0x51a),0x5555aaaa);	/* might break older TOSes <1.02? */
#if SMALL
		SM_UB(MEM(0x424),10);
		SM_UL(MEM(0x42e),0x400000);
		SM_UL(MEM(0x436),0x3f8000);
#else
		SM_UB(MEM(0x424),15);
		SM_UL(MEM(0x42e),0xe00000);
		SM_UL(MEM(0x436),0xdf8000);
#endif
		SM_UL(MEM(0x4ba),16000);
	}
}

void exit_mem()
{
	FreeMem((char *)mem);
}

void mc68000_reset(void)
{
	memset((char *)dreg,0,17*4);
	sr = 0x2700;
	SM_UL(ADDR(0),LM_UL(MEM(TOSSTART)));
	SM_UL(ADDR(4),LM_UL(MEM(TOSSTART+4)));
	SP = LM_UL(ADDR(0));
	pc = LM_UL(ADDR(4));

	pc = TOSSTART;
}
