/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 */
#include "defs.h"
#include "cpu.h"
#include "debug.h"
#include "main.h"
#include "prof.h"
#include "tosdefs.h"
#include <signal.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>

#define IGNORE_S(_addr) void STORE_B_ ## _addr (B v) {}
#define UNDEF_L(_addr) B LOAD_B_ ## _addr (void) { return 0xff; }
#define ZERO(_addr) B LOAD_B_ ## _addr (void) { return 0; }

volatile UW mfp_IERA, mfp_TADR=256, mfp_TACR, mfp_TADR_reload=256,
	mfp_IMRA;
static double timer_a_base;
static struct timeval last_time;
extern int shiftmod;

B LOAD_B_fffa01(void)
{
	if (gr_mode == MODE_MONO)
	{
		return LM_B(MEM(0xfffa01)) & 0x7f;
	}
	return  LM_B(MEM(0xfffa01)) | 0x80;
}

B LOAD_B_fffa03(void)
{
	return /* TODO */ 1;
}

B LOAD_B_fffa05(void)
{
	return /* TODO */ 1;
}

B LOAD_B_fffa07(void)
{
	return mfp_IERA;
}

B LOAD_B_fffa09(void)
{
	return LM_B(MEM(0xfffa09));
}

B LOAD_B_fffa0b(void)
{
#if STE
	return 0x20;
#else
	static B ipra;
	ipra ^= 0xff;
	return ipra;
#endif
}

B LOAD_B_fffa0d(void)
{
	return /* TODO */ 1;
}

B LOAD_B_fffa0f(void)
{
	return /* TODO */ 1;
}

B LOAD_B_fffa11(void)
{
#if 0
	return LM_B(MEM(0xfffa01));
#else
	return LM_B(MEM(0xfffa11));
#endif
}

B LOAD_B_fffa13(void)
{
	return mfp_IMRA;
}

B LOAD_B_fffa15(void)
{
	return /* TODO */ 1;
}

B LOAD_B_fffa17(void)
{
	return /* TODO */ 1;
}

B LOAD_B_fffa19(void)
{
	return mfp_TACR;
}

B LOAD_B_fffa1b(void)
{
	return /* TODO */ 1;
}

B LOAD_B_fffa1d(void)
{
	return /* TODO */ 1;
}

B LOAD_B_fffa1f(void)
{
#if TIMER_A
	return mfp_TADR;
#else
	return mfp_TADR--;
#endif
}

B LOAD_B_fffa21(void)
{ 
	return /* TODO */ 1;
}

B LOAD_B_fffa23(void)
{
	return rand() % 0xff;
}

B LOAD_B_fffa25(void)
{
	return /* TODO */ LM_B(MEM(0xfffa25));
}

B LOAD_B_fffa27(void)
{
	return /* TODO */ 1;
}

B LOAD_B_fffa29(void)
{
	return /* TODO */ 1;
}

B LOAD_B_fffa2b(void)
{
	return LM_B(MEM(0xfffa2b));
}

B LOAD_B_fffa2d(void)
{
	return 0x80|1;
}

B LOAD_B_fffa2f(void)
{
	B x;
	x = read_serial();
#if 0
	fprintf (stderr, "reading UDR -> %d\n", x&0xff);
#endif
	if (!serial_stat())
	{
		SM_UB(MEM(0xfffa2b),LM_UB(MEM(0xfffa2b)) & 0x7f);
	}
#if 1
	else flags |= F_RCV_FULL;
#endif
	return x;
}

UNDEF_L(fffa31)
IGNORE_S(fffa31)

void STORE_B_fffa01(B v)
{
	/* TODO */
}

void STORE_B_fffa03(B v)
{
	/* TODO */
}

void STORE_B_fffa05(B v)
{
	/* TODO */
}

struct { int mode; int div; } mode_A_tab[]=
{{0,1},
 {2,4},
 {2,10},
 {2,16},
 {2,50},
 {2,64},
 {2,100},
 {2,200},
 {1,1},
 {3,4},
 {3,10},
 {3,16},
 {3,50},
 {3,64},
 {3,100},
 {3,200},
};

char *mode_strings[] = {"stopped","event","delay","pulse"};

void STORE_B_fffa07(B v)
{
#if TIMER_A
	UB x=v^mfp_IERA;
	/* TODO */
	if (x & v & 0x20)
	{
		float hz=2457600.0 / mode_A_tab[mfp_TACR&15].div;
		fprintf(stderr,"Timer A Mode: %s, %f (%f) Hz\n", mode_strings[mode_A_tab[mfp_TACR&15].mode], hz,
			hz / mfp_TADR);
		if (mode_A_tab[mfp_TACR&15].mode == 2)
		{
			flags |= F_TIMERA_ON;
			timer_a_base = 1000000.0/hz;
			gettimeofday(&last_time,NULL);
		}
	}
	else if ((x & 0x20) && !(v & 0x20))
	{
		fprintf(stderr,"Timer A switched off\n");
		flags &= ~F_TIMERA_ON;
	}
	mfp_IERA=v;
#endif
}

void STORE_B_fffa09(B v)
{
	SM_B(MEM(0xfffa09),v);
}

void STORE_B_fffa0b(B v)
{
	/* TODO */
}

void STORE_B_fffa0d(B v)
{
	/* TODO */
}

void STORE_B_fffa0f(B v)
{
	/* TODO */
}

void STORE_B_fffa11(B v)
{
	B x = LM_B(MEM(0xfffa11));
	x &= v;
	SM_B(MEM(0xfffa11),x);
}

void STORE_B_fffa13(B v)
{
	mfp_IMRA = v&0xff;
}

void STORE_B_fffa15(B v)
{
	/* TODO */
}

void STORE_B_fffa17(B v)
{
	/* TODO */
}

void STORE_B_fffa19(B v)
{
	mfp_TACR = v;
}

void STORE_B_fffa1b(B v)
{
	/* TODO */
}

void STORE_B_fffa1d(B v)
{
	/* TODO */
}

void STORE_B_fffa1f(B v)
{
#if TIMER_A
	mfp_TADR_reload = (v == 0? 256:v);
	if ((flags & F_TIMERA_ON) == 0)
		mfp_TADR = mfp_TADR_reload;
#endif
}

void STORE_B_fffa21(B v)
{
	/* TODO */
}

void STORE_B_fffa23(B v)
{
	SM_B(MEM(0xfffa23),v);
	/* TODO */
}

void STORE_B_fffa25(B v)
{
	SM_B(MEM(0xfffa25),v);
	/* TODO */
}

void STORE_B_fffa27(B v)
{
	/* TODO */
}

void STORE_B_fffa29(B v)
{
	/* TODO */
}

void STORE_B_fffa2b(B v)
{
	/* TODO */
}

void STORE_B_fffa2d(B v)
{
	/* TODO */
}

void STORE_B_fffa2f(B v)
{
	char c = v;
#if 0
	fprintf (stderr, "Writing UDR: %d\n", v&0xff);
#endif
	write_serial(v);
}

#if TIMER_A

long mfp_get_count (void)
{
	long d;
	double decr;
	long r=0;
	struct timeval t;
	gettimeofday(&t,NULL);
	d = 1000000 * (t.tv_sec-last_time.tv_sec) + t.tv_usec - last_time.tv_usec;
#if 0
	fprintf(stderr,"Delay: %ld\n",d);
#endif
	last_time = t;
	if (d > timer_a_base)
	{
		decr = d/timer_a_base;
		if (decr > mfp_TADR)
		{
			r = 1;
			decr -= mfp_TADR;
			r += decr / mfp_TADR_reload;
			mfp_TADR = (long)decr % mfp_TADR_reload;
		}
		else mfp_TADR -= decr;
#if 0
	fprintf(stderr,"diff/base = %ld/%ld, decr = %d, %d interrupts\n",
			d,timer_a_base, decr, r);
#endif
	}
	return (mfp_IMRA & 0x20) ? r : 0;
}

#endif /* TIMER_A */
