/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 */
#include "defs.h"
#include "mem.h"
#include "screen.h"
#include <stdio.h>
#include <string.h>

static struct {char x[4];} Ztab[] = {
#include "tab8.c"
};

int scr_width,scr_height,scr_planes;
int depth;
int shift_mod;
extern char mapcol[16];

/* Convert an image in ST format (4 interleaved bitplanes) to ZPixmap format
 *
 * TODO: update the video address counter? (some programs wait for it...)
 * Solution: better return random values when the VAC registers are read
 *
 * This code assumes short = 16 bit and int = 32 bit and aligned data!
 */

int chunk_age[MAX_CHUNKS];
static unsigned long chunk_sum[MAX_CHUNKS];
#if IS_BIG_ENDIAN
#define LL(_x) *(_x)
#define SL(_x,_v) *(_x)=_v
#else
/* slow ... :-( */
#define LL(_x) LM_L(_x)	
#define SL(_x,_v) *(_x)=_v
/* #define SL(_x,_v) SM_L(_x,_v) */
#endif
void st16c_to_z (char *st, char *data)
{
	int *y, lo, hi, i, idx, l, q, j,ch;
	unsigned short w1,w2,w3,w4,*x;
	unsigned int a1, a2, r1, r2, *r;
	int f0;
	int *ztab = (int *)Ztab;
	int chunk_bytes=(scr_width/2)*CHUNK_LINES;
	r = (unsigned int *)st;
	y = (int *)data;
	f0 = (scr_width/16)*CHUNK_LINES;
	for (ch=0; ch<scr_height/CHUNK_LINES; ch++)
	{
		if (chunky && ++chunk_age[ch] < MAX_AGE)
		{
			unsigned long sum=0;
			int p;
			unsigned long *chp = (unsigned long *)r;
			for (p=0; p<chunk_bytes/sizeof(*chp); p++)
				sum += chp[p];
			if (sum == chunk_sum[ch])
			{
				y += 4*f0;
				r += 2*f0;
				continue;
			}
			chunk_sum[ch]=sum;
		}
		chunk_age[ch]=1;
		for (j=0; j<f0; j++)
		{
			r1 = LL(r);
			r2 = LL(r+1);
			r+=2;
			a1 = (r1 & 0xf0f0f0f); 
			w1 = (a1 << 4) | (a1 >> 16);
			a2 = (r2 & 0xf0f0f0f);
			w2 = (a2 << 4) | (a2 >> 16);
			SL(y+3,ztab[w1&0xff] + 4 * ztab[w2&0xff]);
			SL(y+1,ztab[w1>>8] + 4 * ztab[w2>>8]);
			a1 = (r1 & 0xf0f0f0f0); 
			w1 = (a1 >> 20) | a1;
			a2 = (r2 & 0xf0f0f0f0);
			w2 = (a2 >> 20) | a2;
			SL(y+2,ztab[w1&0xff] + 4 * ztab[w2&0xff]);
			SL(y,ztab[w1>>8] + 4 * ztab[w2>>8]);
			y += 4;
		}
		if (!priv_cmap)
		{
			int p=CHUNK_LINES*ch*scr_width;
			for (j=0; j<scr_width*CHUNK_LINES; j++)
			{
				data[p+j]=mapcol[data[p+j]];
			}
		}
	}
}

/* Convert an image in 2 interleaved bitplane format to ZPixmap with 400
 * lines. An optimizing compiler is a good idea ;-) (CSE)
 */
void st4c_to_z (char *st, char *data)
{
	int *r = (int *)st, r1, r2, ch;
	int i, j, k;
	int chunk_bytes=(scr_width/4)*CHUNK_LINES;
	for (ch=0; ch<scr_height/CHUNK_LINES; ch++)
	{
		if (chunky && ++chunk_age[ch] < MAX_AGE)
		{
			unsigned long sum=0;
			int p;
			unsigned long *chp = (unsigned long *)r;
			for (p=0; p<chunk_bytes/sizeof(*chp); p++)
				sum += chp[p];
			if (sum == chunk_sum[ch])
			{
				data += scr_width*2*CHUNK_LINES;
				r += chunk_bytes/4;
				continue;
			}
			chunk_sum[ch]=sum;
		}
		chunk_age[ch]=1;

		for (k=0; k<CHUNK_LINES; k++, data += scr_width)
		for (i=0; i<scr_width/16; i++)
		{
			r1 = LL(r);
			r2 = ((r1>>16)&0x5555)|((r1<<1)&0xaaaa);
			if (priv_cmap)
			{
				for (j=0; j<8; j++)
				{
					*(data+15-2*j) = r2 & 3;
					*(data+scr_width+15-2*j) = r2 & 3;
					r2 >>= 2; 
				}
				r2 = ((r1>>17)&0x5555)|(r1&0xaaaa);
				for (j=0; j<8; j++)
				{
					*(data+14-2*j) = r2 & 3;
					*(data+scr_width+14-2*j) = r2 & 3;
					r2 >>= 2;
				}
			}
			else
			{
				for (j=0; j<8; j++)
				{
					*(data+15-2*j) = mapcol[r2 & 3];
					*(data+scr_width+15-2*j) = mapcol[r2 & 3];
					r2 >>= 2;
				}
				r2 = ((r1>>17)&0x5555)|(r1&0xaaaa);
				for (j=0; j<8; j++)
				{
					*(data+14-2*j) = mapcol[r2 & 3];
					*(data+scr_width+14-2*j) = mapcol[r2 & 3];
					r2 >>= 2;
				}
			}
			data += 16;
			r++;
		}
	}
}

/* Convert an image in 2 interleaved bitplane format to ZPixmap with 200
 * lines. An optimizing compiler is a good idea ;-) (CSE)
 */
void st4c_to_z200 (char *st, char *data)
{
	int *r = (int *)st, r1, r2, ch;
	int i, j, k;
	int chunk_bytes=(scr_width/4)*CHUNK_LINES;
	for (ch=0; ch<scr_height/CHUNK_LINES; ch++)
	{
		if (chunky && ++chunk_age[ch] < MAX_AGE)
		{
			unsigned long sum=0;
			int p;
			unsigned long *chp = (unsigned long *)r;
			for (p=0; p<chunk_bytes/sizeof(*chp); p++)
				sum += chp[p];
			if (sum == chunk_sum[ch])
			{
				data += scr_width*CHUNK_LINES;
				r += chunk_bytes/4;
				continue;
			}
			chunk_sum[ch]=sum;
		}
		chunk_age[ch]=1;

		for (k=0; k<CHUNK_LINES; k++)
		for (i=0; i<scr_width/16; i++,r++,data+=16)
		{
			r1 = LL(r);
			r2 = ((r1>>16)&0x5555)|((r1<<1)&0xaaaa);
			data[15] = r2 & 3; r2 >>= 2;
			data[13] = r2 & 3; r2 >>= 2;
			data[11] = r2 & 3; r2 >>= 2;
			data[9] = r2 & 3; r2 >>= 2;
			data[7] = r2 & 3; r2 >>= 2;
			data[5] = r2 & 3; r2 >>= 2;
			data[3] = r2 & 3; r2 >>= 2;
			data[1] = r2 & 3; r2 >>= 2;
			r2 >>= 2;
			r2 = ((r1>>17)&0x5555)|(r1&0xaaaa);
			data[14] = r2 & 3; r2 >>= 2;
			data[12] = r2 & 3; r2 >>= 2;
			data[10] = r2 & 3; r2 >>= 2;
			data[8] = r2 & 3; r2 >>= 2;
			data[6] = r2 & 3; r2 >>= 2;
			data[4] = r2 & 3; r2 >>= 2;
			data[2] = r2 & 3; r2 >>= 2;
			data[0] = r2 & 3; r2 >>= 2;
		}
	}
}

unsigned long ExpTabHi[0x200];

inline asm void convert18(register double *dest,register unsigned char *src,register unsigned long *TabHi, register unsigned long todo, register double work1, register char work2, register char work3)
{	
		// by Tinic Urou <5uro@informatik.uni-hamburg.de>
		//
		// 0.6 instructions per pixel.
		//
		// possible enhancements: first fill all registers
		// then write the bytes for optimal cache usage.

		subi	dest,dest,8 	  // init pointers for pointer updating
		subi	src,src,1

		li		work3,3			  // initialise decrement counter CTR
		srw		todo,todo,work3
		mtctr	todo			
		
@loop	
		lbzu	work2,1(src)	  // get a source byte, and update source pointer
		slw		work2,work2,work3 // shift it to double word length

		lfdx	work1,TabHi,work2 // get 8 bytes
		stfdu	work1,8(dest)	  // write them to the destination buffer
	
		bdnz	@loop			  // loop until CTR is 0

		blr
};

void stmono_to_xy (char *data, char *st)
{
	int i;
	int bytes_per_chunk,bc;
	unsigned long *r=(unsigned long *)st;
	
	bytes_per_chunk = CHUNK_LINES*scr_width/8;
	bc = bytes_per_chunk * 8;
	
	if (!chunky)
	{
		convert18((double *)data,(unsigned char *)st,ExpTabHi,scr_height*scr_width,0,0,0);
		return;
	}

	for (i=0; i<scr_height/CHUNK_LINES; i++)
	{
		if (++chunk_age[i] < MAX_AGE)
		{
			unsigned long sum=0;
			int p;
			unsigned long *chp = r;
			for (p=0; p<bytes_per_chunk/sizeof(*chp); p++) sum += chp[p];
			if (sum == chunk_sum[i])
			{
				r += bytes_per_chunk/sizeof(*r);
				data += bc;
				continue;
			}
			chunk_sum[i]=sum;
		}
		chunk_age[i]=1;
		convert18((double *)data,(unsigned char *)r,ExpTabHi,bc,0,0,0);		
		r += bytes_per_chunk/sizeof(*r);
		data += bc;
	}
}
