/* This file is part of STonX, the Atari ST Emulator for Unix/X
 * ============================================================
 * STonX is free software and comes with NO WARRANTY - read the file
 * COPYING for details
 *
 * BUGS:
 * - vro_cpyfm probabyl doesn't work properly when fancy modes are used for
 *   screen->memory copying
 * - user-defined fill patterns are global (?)
 *
 * - need to 1) debug and 2) optimize the raster ops (in that order)
 *
 * - should use own vqt_attributes() if custom screen size looks better with
 *   larger font (really? maybe I just need to patch the appropriate variables)
 */
#include "defs.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mem.h"
#include "missing.h"
#include "screen.h"
#include "xlib_vdi.h"

#if XBUFFER
#define XBUF(_x) _x
#else
#define XBUF(_x)
#endif

#define VDEBUG 0

UL abase,fontbase;
int vdi_ok=0,linea_ok=0;
extern	int depth,vdi,fix_screen,priv_cmap;
static int vdi_planes;

#if VDEBUG
#define T(_x) fprintf (stderr, #_x "\n")
#else
#define T(_x) 
#endif

#define IGNORE(_x) static void _x(void){T(unimpl_ ## _x);}

extern int scr_width,scr_height,scr_planes;
#define vdi_w scr_width
#define vdi_h scr_height
int vdi_done = 0;
extern volatile UL vbase;

/*
	TODO:

Opcode	Function

	1		Open Workstation
	2		Close Workstation
	3		Clear Workstation
	4		Update Workstation
	5		Escape Functions
		1	Inquire addressable character cells
		2	Exit alpha mode
		3	Enter alpha mode
		4	Cursor up
		5	Cursor down
		6	Cursor right
		7	Cursor left
		8	Home Cursor
		9	Erase to end of screen
	   10	Erase to end of line
	   11	Direct cursor address
	   12	Output cursor addressable text
	   15	Inquire current alpha cursor address
	   18	Place graphic cursor
	   19	Remove last graphic cursor
	6		Polyline
	7		Polymarker
	8		Text
	9		Filled area
   11		GDP
		1	Bar
		2	Arc
		3	Pie
		4	Circle
		5	Ellipse
		6	Elliptical arc
		7	Elliptical pie
		8	Rounded rectangle
		9	Filled rounded rectangle
	   10	Justified graphics text
	12		Set character height, absolute mode
	14		Set color representation
	15		Set polyline linetype
	17		Set polyline color index
	18		Set polymarker type
	20		Set polymarker color index
	21		Set text face
	22		Set graphics text color index
	23		Set fill interior style
	24		Set fill style index
	25		Set fill color index
	26		Inquire color representation
	28		Input locator
	31		Input string
	32		Set writing mode
	33		Set input mode
	35		Inquire current polyline attributes
	36		Inquire current polymarker attributes
	37		Inquire current fill area attributes
	38		Inquire current graphic text attributes
	39		Set graphic text alignment
   100		Open virtual screen workstation
   101		Close virtual screen workstation
   102		Extended inquire function
   104		Set fill perimeter visibility
   106		Set graphic text special effects
   107		Set character cell height, points mode
   108		Set polyline end styles
   109		Copy raster, opaque
   110		Transform form
   111		Set mouse form
   112		Set used-defined fill pattern
   113		Set user-defined linestyle
   114		Fill rectangle
   115		Inquire input mode
   116		Inquire text extent
   117		Inquire character cell width
   118		Exchange timer interrupt vector
   121		Copy raster, transparent
   122		Show cursor
   123		Hide cursor
   124		Sample mouse button state
   125		Exchange button change vector
   126		Exchange mouse movement vector
   127		Exchange cursor change vector
   128		Sample keyboard state information
   129		Set clipping rectangle
   130		Inquire face name and index
   131		Inquire current face information

*/
#define MAX_VWK 100
#define MAX_POINTS 1000
#define MAX_VDI_COLS	256

static int vdi_maptab[MAX_VDI_COLS];
static int vdi_maptab16[]={
0,15,1,2,4,6,3,5,7,8,9,10,12,14,11,13};

static UW work_out_buf[128];

static UL control, intin, ptsin, intout, ptsout;
#define CONTRL(_x) MEM(control+2*(_x))
#define INTIN(_x) MEM(intin+2*(_x))
#define PTSIN(_x) MEM(ptsin+2*(_x))
#define INTOUT(_x) MEM(intout+2*(_x))
#define PTSOUT(_x) MEM(ptsout+2*(_x))
#define SW(_x,_v) SM_W(_x,_v)

void Vdi (UL as)
{
	UW osr = sr;
	W c;
	UL pblock;
	UNSET_Z();
	if (!vdi || DREG(0) != 115) return;
	pblock = DREG(1);
	sr |= 0x700;
	control	= LM_UL(MEM(pblock));
	intin 	= LM_UL(MEM(pblock+4));
	ptsin 	= LM_UL(MEM(pblock+8));
	intout 	= LM_UL(MEM(pblock+12));
	ptsout 	= LM_UL(MEM(pblock+16));
	c = LM_W(CONTRL(0));
#if VDEBUG
	if (c != 128 && c != 31 && c != 33)fprintf (stderr, "VDI %d: ", c);
#endif
	vdi_done=0;
//	if (c >= 1 && c <= N_VDIFUNC)
//		vdifunc[c-1]();
	sr = osr;
	if (vdi_done) SET_Z();
}

void linea_init (void)
{
#if 0
	if (scr_width != 640 || scr_height != 400)
	{
	fprintf (stderr,"Line-A Initialization\n");
	SM_UL(MEM(0xfa1200),LM_UL(MEM(0x28)));
	SM_UL(MEM(0x28),0xfa1204);
	}
#endif
}

void init_vdi (void)
{
#if 0
	fprintf(stderr,"Grabbing Trap #2...\n");
#endif
	SM_UL(MEM(0xfa1000),LM_UL(MEM(0x88)));
	SM_UL(MEM(0x88),0xfa1004);
#if 0
	if (fix_screen)
	{
		SM_UL(MEM(0x44e),0x300000);
		SM_UL(MEM(0x42e),0x300000);
		SM_UL(MEM(0x436),0x300000);
		vbase=0x300000;
	}
#endif
	vdi_ok=1;
}

#define CHAR_HEIGHT() (vdi && scr_height >= 400 || scr_planes==1 ? 16 : 8)

void fix_vars (void)
{
	UW *a;
	a=MEM(abase);
	SM_W(a+1,(scr_width*scr_planes)/8);
	SM_W(a-1,(scr_width*scr_planes)/8);
	SM_W(a-23,CHAR_HEIGHT());
	SM_W(a-22,scr_width/8-1);
	SM_W(a-21,scr_height/CHAR_HEIGHT()-1);
	SM_W(a-20,((scr_width*scr_planes)/8)*CHAR_HEIGHT());
	SM_W(a-6,scr_width);
	SM_W(a-2,scr_height);
#if 0
	fprintf(stderr,"FBASE=%08x, CUR_FONT=%08x, DEF_FONT=%08x, [%d,%d,%d,%d]\n",
		LM_UL(a+0x54/2),LM_UL(a-0x38a/2),LM_UL(a-0x1cc/2),
		LM_W(a-0x1f2/2),LM_W(a-0x1f0/2),LM_W(a-0x1ee/2),LM_W(a-0x1ec/2));
	if (CHAR_HEIGHT() == 16)
	{
		SM_UL(a-0x1cc/2,LM_UL(MEM(fontbase+8)));
		SM_UL(a-0x38a/2,LM_UL(MEM(fontbase+8)));
	}
	fprintf(stderr,"-> FBASE=%08x, CUR_FONT=%08x, DEF_FONT=%08x\n",
		LM_UL(a+0x54/2),LM_UL(a-0x38a/2),LM_UL(a-0x1cc/2));
#endif
}

void vdi_post (UL as)
{
	UW *a;
	UL pblock = DREG(1);
	control	= LM_UL(MEM(pblock));
	intin 	= LM_UL(MEM(pblock+4));
	ptsin 	= LM_UL(MEM(pblock+8));
	intout 	= LM_UL(MEM(pblock+12));
	ptsout 	= LM_UL(MEM(pblock+16));
	if (LM_UW(CONTRL(0)) != 1 && LM_UW(CONTRL(0)) != 100)
	{
		return;
	}
#if 0
	fprintf(stderr,"abase=$%lx (%d), screen=%lx\n",abase,fix_screen,LM_UL(MEM(0x44e)));
#endif
	a=MEM(abase);
	if (fix_screen)
	{
		SM_W(INTOUT(0),vdi_w-1);
		SM_W(INTOUT(1),vdi_h-1);
		SM_W(a-0x15a,vdi_w-1);
		SM_W(a-0x159,vdi_h-1);
#if 0
		SM_B(MEM(0x44c),3);
		fprintf(stderr,"Fixing Line-A variables...\n");
#endif
		fix_vars();
	}
	vdi_planes = LM_W(a);

	/* fix the lookup table */
	if (LM_UW(CONTRL(0)) == 1)
	{
		int i;
		for (i=0; i<MAX_VDI_COLS; i++)
			vdi_maptab[i] = (1<<vdi_planes)-1;
		for (i=0; i<(1<<vdi_planes); i++)
			vdi_maptab[i] = vdi_maptab16[i] & ((1<<vdi_planes)-1);
	}

	memcpy(&work_out_buf[0],INTOUT(0),90);
	memcpy(&work_out_buf[45],PTSOUT(0),24);
}

void linea_post (UL as)
{
	UW *a=MEM(DREG(0));
	abase=DREG(0);
	if (fix_screen) fix_vars();
}

void Init_Linea (UL as)
{
	UW *a;
	abase=DREG(0);
	fontbase=AREG(1);
	fprintf(stderr,"Initializing Line-A... (abase=%04lx)\n",
		abase);
	a = (UW *)MEM(abase);
	if (fix_screen) fix_vars();
}

